use std::sync::mpsc;
use std::thread;

pub mod models;
mod datastore;
mod nfqueue;
mod tshark;

pub use models::FirewallClient;
pub use models::Policy;
pub use models::Rule;
use models::MatchStatus;

pub mod lib;
pub mod schema;

use datastore::run as datastore_run;
use self::nfqueue::run as nfqueue_run;
use tshark::run as tshark_run;
use tshark::shutdown as tshark_shutdown;

pub fn run(advertisement_get_client_rx_receiver: mpsc::Receiver<models::FirewallClient>, advertisement_get_client_tx_sender: mpsc::Sender<Option<models::FirewallClient>>,
    advertisement_put_client_receiver: mpsc::Receiver<models::FirewallClient>, advertisement_get_policy_rx_receiver: mpsc::Receiver<models::Policy>,
    advertisement_get_policy_tx_sender: mpsc::Sender<Option<models::Policy>>, advertisement_put_policy_receiver: mpsc::Receiver<models::Policy>) {
    let (nfqueue_tx_sender, nfqueue_tx_receiver) = mpsc::channel();
    let (nfqueue_rx_sender, nfqueue_rx_receiver) = mpsc::channel();
    let (tshark_sender, tshark_receiver) = mpsc::channel();
    let datastore_thread = thread::spawn(|| {
        datastore_run(nfqueue_tx_receiver, nfqueue_rx_sender, tshark_receiver,
            advertisement_get_client_rx_receiver, advertisement_get_client_tx_sender, advertisement_put_client_receiver,
            advertisement_get_policy_rx_receiver, advertisement_get_policy_tx_sender, advertisement_put_policy_receiver
        );
    });
    let nfqueue_thread = thread::spawn(|| nfqueue_run(nfqueue_tx_sender, nfqueue_rx_receiver));
    let tshark_thread = thread::spawn(|| tshark_run(tshark_sender));
    datastore_thread.join();
    tshark_thread.join();
    nfqueue_thread.join();
}

pub fn shutdown() {
    tshark_shutdown();
}
