#[macro_use]
extern crate diesel;
extern crate dotenv;
#[macro_use]
extern crate log;

use std::process::exit;
use std::sync::mpsc;
use std::thread;

mod advertisement;
use advertisement::run as advertisement_run;
use advertisement::shutdown as advertisement_shutdown;
mod firewall;
use firewall::run as firewall_run;
use firewall::shutdown as firewall_shutdown;



extern "C" {
    fn signal(sig: u32, cb: extern fn(u32)) -> fn(u32);
}

extern fn shutdown(_signum: u32) {
    println!("Shutting down");
    advertisement_shutdown();
    firewall_shutdown();
}

fn main() {
    env_logger::init();
    unsafe {
        signal(15, shutdown);
        signal(2, shutdown);
    }
    let (advertisement_get_client_rx_sender, advertisement_get_client_rx_receiver) = mpsc::channel();
    let (advertisement_get_client_tx_sender, advertisement_get_client_tx_receiver) = mpsc::channel();
    let (advertisement_put_client_sender, advertisement_put_client_receiver) = mpsc::channel();
    let (advertisement_get_policy_rx_sender, advertisement_get_policy_rx_receiver) = mpsc::channel();
    let (advertisement_get_policy_tx_sender, advertisement_get_policy_tx_receiver) = mpsc::channel();
    let (advertisement_put_policy_sender, advertisement_put_policy_receiver) = mpsc::channel();
    let firewall_thread = thread::spawn(|| {
        firewall_run(advertisement_get_client_rx_receiver, advertisement_get_client_tx_sender, advertisement_put_client_receiver,
            advertisement_get_policy_rx_receiver, advertisement_get_policy_tx_sender, advertisement_put_policy_receiver);
    });
    let advertisement_thread = thread::spawn(|| {
        advertisement_run(advertisement_get_client_rx_sender, advertisement_get_client_tx_receiver, advertisement_put_client_sender,
            advertisement_get_policy_rx_sender, advertisement_get_policy_tx_receiver, advertisement_put_policy_sender);
    });
    advertisement_thread.join();
    firewall_thread.join();
    exit(0);
}

#[cfg(test)]
mod tests {
    // Note this useful idiom: importing names from outer (for mod tests) scope.
    use super::*;

    #[test]
    fn test_main() {
        assert!(true);
    }

}
