import ast
import base64
import dns.resolver
from http.server import BaseHTTPRequestHandler, HTTPServer
import os
import pytest
import requests
from shutil import copyfile
import socket
import struct
import subprocess
from time import sleep
from Crypto.PublicKey import RSA
from Crypto import Random
from Crypto.Hash import SHA256
from Crypto.Signature import PKCS1_v1_5


PRIVATE_KEY = b'-----BEGIN RSA PRIVATE KEY-----\n' \
    b'MIIEpgIBAAKCAQEAvQMKf5Ta1RBSW/ihfsPkToHDNfkW8pDw0merXxQhYprNUqDM\n' \
    b'9Qow9WGpW8xhrHclEHv28ORV1mXVBLRXxTR1jQHzkOQZkMFNdXgjo9rTivF91qDW\n' \
    b'WC4vrt71w5eLzk0JuY1s1Lah39M4WVK87vxEzCveARorcN5+ydxnUNv0Cl2cj1n7\n' \
    b'REc0JmQx3NmEVd5avSsG0ExOatM3mO6WjhbIv1/lSGYD+KnEXQCcTSP/gIY0OwaY\n' \
    b'xk5hp3//8k8y4E5/+9SiNuDFilW0Tpq2yKRBTyEcSDTTF4s6EmBK575WbfusEtdZ\n' \
    b'LN0hFB3E3bK4rGwoqv447D/ETcocqMXYJaJejQIDAQABAoIBAQCZqjxBxyI2MIPU\n' \
    b'1AlvxxD5JpQvcChQ4rd3R+NSdEpEITimYabf3P46sMCAqmQKU/8lLLNx2yFSIg4S\n' \
    b'OhBVerXLAvgjJt8GCbHQu/C1yTVM89JryO0aWcpYcFn8P1vhSE6MtUbylNx73Pwl\n' \
    b'2pdR8HR6TXu7B6yjobAQqwiv+3KnbVceQsQsjxLbHRAAkg7W0BUEKGRBGZPwBoyc\n' \
    b'mgDoeapo0ACwJbbokePZ+omp/3OvVpo78TzO751TCh8mAIvIE9oUwbmPfHt+dB/t\n' \
    b'kQeowo6Ouv3EAm+U8whc/ZVryMnuR72FFF2xINjoqZmcLBCT6UtxO8Opmk3A6zCd\n' \
    b'CbnTJLmdAoGBAO2CuDIfLRYV5Xs1H576kwUCGwpaDB5K2AF8Ktxw2SkFT2mG3vVv\n' \
    b'9vV1/IqJ3AwsGTqZvzTc+L/9dokWVzEvRU13/uTzlnuQEu90/YnPEKnOh5bULAxC\n' \
    b'w2HSU09Ky8GJNUsgo9/vLHCiPiBNFL1NAJB2M+1/xk+pgu1Via6jj7ezAoGBAMu5\n' \
    b'zdaLMQATypk+EzL958ltVamECUjKGD0Hyf1qlszMK99o6sYHiz1G5Zv1T97Fe+o/\n' \
    b'W0TBwYtMx7X7pdF1dY/lgFRE6zVk/WHZs14ZbQrEHOdiCOw76TAdKLzbX83DDV1Q\n' \
    b'iZIEOl96Zkd7+QrFnzqn58bf0kYcngIgR4VpznC/AoGBAID+wB+owwn9tncmMRhd\n' \
    b'K0BLFniW04IYUBzuXTEAsVEQK0M7FxjiSJSkEhcELNkHkmQ7YJfLc+aCOMDXu19h\n' \
    b'NxiOFYgvrAlMBJy9Q353UauIGigW8B1wnUEhxFGKZAOnC6TUA2z6L360/1z3wE3X\n' \
    b'5niPWkChZwmZHRzdmjuZZdL1AoGBAIfkCVQ5hlbZ3MM5L89MpAatAn3PT+4Yt/BF\n' \
    b'uMtsBOjYq88bCbdMIumd19ye1kbvjcdjuXp0j6heb7yI/kNoSSHR4TbUvdvg4E8b\n' \
    b'ytRSu3R5+hmxP+hr0hDA8oJ/BjmXCNwcs5cNcO5tE7pNDILN7XdPtn/mnjkcL52V\n' \
    b'cH+iYwy7AoGBANiKz9Xjjn8q8q2IsAQrrGBrgPgsTM99j8U+Nnp88gfM8/mrj0rV\n' \
    b'svlVtU+UKcXhrkRbbtw4If7V92/r6FWBQntwhrYr69e56MbqKOFC+CGZRmZ16Do1\n' \
    b'Vp3IUecBxc0/XA5DonvnkzwsYpY66oXNZJtv0V1c0cguegVnVfR9XxIK\n' \
    b'-----END RSA PRIVATE KEY-----'

class DummyRequestHandler(BaseHTTPRequestHandler):
    def do_POST(self):
        # TODO
        pass

class DummyServer(HTTPServer):
    timeout = 5

    def handle_timeout(self):
        assert False

def start_firewall():
    assert subprocess.call('sh/start-firewall.sh') == 0

def stop_firewall():
    assert subprocess.call('sh/stop-firewall.sh') == 0

def reset_firewall():
    assert subprocess.call('sh/reset-firewall.sh') == 0

def send_cloudflare_dns_request():
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ['1.1.1.1']
    return resolver.query('example.com')

def send_cloudflare_ipv6_dns_request():
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ['2606:4700:4700::1111']
    return resolver.query('example.com')

def send_quad9_dns_request():
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ['9.9.9.9']
    return resolver.query('example.com')

def get_cloudflare_dns_homepage():
    res = requests.get("http://1.1.1.1/", timeout=2)
    assert res.status_code == 200

def send_advertisement(advertisement_filename):
    # Get Message from file
    message_readable = 'ADVERTISE * HTTP/1.1'
    advertisement_file = open(advertisement_filename, 'rb')
    body = advertisement_file.read()
    advertisement_file.close()

    # Create signature for BODY code.
    # This would normally already be done by the manufacturer.
    sha256_hash = SHA256.new()
    sha256_hash.update(body)
    key = RSA.importKey(PRIVATE_KEY)
    signature = PKCS1_v1_5.new(key).sign(sha256_hash)
    # convert UTF-8? encoding to base64
    signature = base64.b64encode(signature)

    # Create encoded message to be sent
    message = bytes(message_readable + '\r\n' + 'Signature: ', 'UTF-8',
                    errors='strict') + signature + \
        bytes('\r\n\r\n' + body.decode() + '\r\n', 'UTF-8', errors='strict')

    multicast_group = ('192.168.90.1', 8888)
    # Create the datagram socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # Set a timeout so the socket does not block indefinitely when trying
    # to receive data.
    sock.settimeout(0.2)
    # Set the time-to-live for messages to 1 so they do not go past the
    # local network segment.
    ttl = struct.pack('b', 1)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
    try:
        # Send data to the multicast group
        sock.sendto(message, multicast_group)
    finally:
        print("closing socket")
        sock.close()

def send_advertisement_bad_signature(advertisement_filename):
    # Get Message from file
    message_readable = 'ADVERTISE * HTTP/1.1'
    advertisement_file = open(advertisement_filename, 'rb')
    body = advertisement_file.read()
    advertisement_file.close()

    # Create signature for BODY code.
    # This would normally already be done by the manufacturer.
    sha256_hash = SHA256.new()
    sha256_hash.update(body)
    key = RSA.importKey(PRIVATE_KEY)
    signature = PKCS1_v1_5.new(key).sign(sha256_hash)
    # convert UTF-8? encoding to base64
    signature = b'4' + base64.b64encode(signature)

    # Create encoded message to be sent
    message = bytes(message_readable + '\r\n' + 'Signature: ', 'UTF-8',
                    errors='strict') + signature + \
        bytes('\r\n\r\n' + body.decode() + '\r\n', 'UTF-8', errors='strict')

    multicast_group = ('192.168.90.1', 8888)
    # Create the datagram socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # Set a timeout so the socket does not block indefinitely when trying
    # to receive data.
    sock.settimeout(0.2)
    # Set the time-to-live for messages to 1 so they do not go past the
    # local network segment.
    ttl = struct.pack('b', 1)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
    try:
        # Send data to the multicast group
        sock.sendto(message, multicast_group)
    finally:
        print("closing socket")
        sock.close()

def put_policy_in_ftp_directory(advertisement_filename, tag):
    copyfile(advertisement_filename, '/srv/ftp/{}/test.xml'.format(tag.replace('.', '/')))

def clear_policy_in_ftp_directory(tag):
    os.remove('/srv/ftp/{}/test.xml'.format(tag.replace('.', '/')))

def test_01():
    """Test that the firewall blocks packets according to policy"""
    reset_firewall()
    start_firewall()
    try:
        send_cloudflare_dns_request()
        send_advertisement('advertisements/block_cloudflare_dns.xml')
        sleep(0.1)
        try:
            send_cloudflare_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
    finally:
        stop_firewall()

def test_02():
    """Test that the firewall updates the policy and passes packets according to policy"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/block_everything.xml')
        sleep(0.1)
        try:
            send_cloudflare_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
        send_advertisement('advertisements/pass_cloudflare_dns.xml')
        sleep(0.1)
        send_cloudflare_dns_request()
    finally:
        stop_firewall()

def test_03():
    """Test that the firewall passes packets from unknown hosts"""
    reset_firewall()
    start_firewall()
    try:
        send_cloudflare_dns_request()
    finally:
        stop_firewall()

def test_04():
    """Test that the firewall defaults to drop packets that do not match a rule"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/no_rules.xml')
        sleep(0.1)
        try:
            send_cloudflare_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
    finally:
        stop_firewall()

def test_05():
    """Test that conditions with multiple selectors work properly"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/pass_cloudflare_dns.xml')
        sleep(0.1)
        send_cloudflare_dns_request()
        try:
            send_quad9_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
        try:
            get_cloudflare_dns_homepage()
            assert False
        except requests.exceptions.Timeout:
            pass
    finally:
        stop_firewall()

def test_06():
    """Test that rules with multiple conditions work properly"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/pass_cloudflare_or_dns.xml')
        sleep(0.1)
        send_cloudflare_dns_request()
        send_quad9_dns_request()
        get_cloudflare_dns_homepage()
    finally:
        stop_firewall()

@pytest.mark.xfail(reason='IPv6 not enabled on the test infrastructure')
def test_07a():
    """Test that the firewall blocks IPv6 packets according to policy"""
    reset_firewall()
    start_firewall()
    try:
        send_cloudflare_ipv6_dns_request()
        send_advertisement('advertisements/block_cloudflare_ipv6_dns.xml')
        sleep(0.1)
        try:
            send_cloudflare_ipv6_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
    finally:
        stop_firewall()

@pytest.mark.xfail(reason='IPv6 not enabled on the test infrastructure')
def test_07b():
    """Test that the firewall updates the policy and passes IPv6 packets according to policy"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/block_everything.xml')
        sleep(0.1)
        try:
            send_cloudflare_ipv6_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
        send_advertisement('advertisements/pass_cloudflare_ipv6_dns.xml')
        sleep(0.1)
        send_cloudflare_ipv6_dns_request()
    finally:
        stop_firewall()

def test_08():
    """Test that notifications work properly"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/notify_cloudflare_dns.xml')
        sleep(0.1)
        server = DummyServer(('0.0.0.0', 8000), DummyRequestHandler)
        send_cloudflare_dns_request()
        server.handle_request()
    finally:
        stop_firewall()

def test_12():
    """Test that the firewall retrieves and applies managed policies"""
    reset_firewall()
    start_firewall()
    try:
        put_policy_in_ftp_directory('advertisements/pass_cloudflare_dns.xml', 'com.phillips.hue')
        send_advertisement('advertisements/tag.xml')
        sleep(0.1)
        send_cloudflare_dns_request
        try:
            send_quad9_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
    finally:
        stop_firewall()
        clear_policy_in_ftp_directory('com.phillips.hue')

def test_15():
    """Test that the firewall rejects advertisements with bad signatures"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/block_everything.xml')
        sleep(0.1)
        try:
            send_cloudflare_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
        send_advertisement_bad_signature('advertisements/pass_cloudflare_dns.xml')
        sleep(0.1)
        try:
            send_cloudflare_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
    finally:
        stop_firewall()

def test_16():
    """Test that the firewall does not update a cached policy if the newly
    advertised policy has the same revision number"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/pass_cloudflare_or_dns.xml')
        sleep(0.1)
        send_cloudflare_dns_request()
        get_cloudflare_dns_homepage()
        send_advertisement('advertisements/block_everything.xml')
        # policy will not be applied because it has the same revision number
        sleep(0.1)
        send_cloudflare_dns_request()
        get_cloudflare_dns_homepage()
    finally:
        stop_firewall()

def test_17():
    """Test that the firewall persists policies across restarts"""
    reset_firewall()
    start_firewall()
    try:
        send_advertisement('advertisements/block_everything.xml')
        sleep(0.1)
    finally:
        stop_firewall()
    start_firewall()
    try:
        try:
            send_cloudflare_dns_request()
            assert False
        except dns.exception.Timeout:
            pass
        send_advertisement('advertisements/pass_cloudflare_dns.xml')
        sleep(0.1)
    finally:
        stop_firewall()
    start_firewall()
    try:
        send_cloudflare_dns_request()
    finally:
        stop_firewall()
