#!/bin/bash

MIDDLEBOX=192.168.90.1
CLIENT=192.168.90.2
SERVER=192.168.7.237

function run_on_middlebox {
	ssh $MIDDLEBOX "$@"
}

function run_on_client {
	ssh $CLIENT "$@"
}

function run_on_server {
	ssh $SERVER "$@"
}

function start_firewall {
	run_on_middlebox iptables -F
	run_on_middlebox 'cd /srv/iot-security-mqp/middlebox && export RUST_LOG=debug && export RUST_BACKTRACE=1 && { nohup /root/.cargo/bin/cargo run >output.txt 2>&1 & echo $! > pid ; disown ; }'
	count=0
	until run_on_middlebox 'grep -q "Running \`target/debug/middlebox\`" /srv/iot-security-mqp/middlebox/output.txt'; do
		((count++))
		[ "$count" -eq 10 ] && return 1
		sleep 3
	done
}

function stop_firewall {
	count=0
	until ! run_on_middlebox 'kill -TERM $(cat /srv/iot-security-mqp/middlebox/pid)'; do
		((count++))
		[ "$count" -eq 3 ] && return 1
		ping -w 1 google.com
		sleep 3
	done
}

function reset_firewall {
	run_on_middlebox 'cd /srv/iot-security-mqp/middlebox && /root/.cargo/bin/diesel migration redo'
}

function setup_iptables {
	run_on_middlebox iptables -F
	run_on_middlebox iptables -A POSTROUTING -o eth1 -t nat -j MASQUERADE
	run_on_middlebox iptables -A FORWARD -i eth0 -t filter -j ACCEPT
}

function teardown_iptables {
	run_on_middlebox iptables -F
}

function send_file_to_server {
	run_on_client cd /srv/iot-security-mqp/tests && python3 HTTP_transfer.py -m send -i large.m4v
	sleep 3
	run_on_server cd /root/iot-security-mqp/tests && python3 HTTP_transfer.py -m receive -i 192.168.90.1
}

function send_file_to_client {
	run_on_server cd /root/iot-security-mqp/tests && python3 HTTP_transfer.py -m send -i test.txt
	sleep 3
	run_on_client cd /srv/iot-security-mqp/tests && python3 HTTP_transfer.py -m receive -i 192.168.7.237
}
