# MQP: IoT Network Security

**Major Qualifying Project for Worcester Polytechnic Institute**

**March 2020** 


## Authors

* **Matthew Hagan** - [SubBassBro](https://github.com/subbassbro)
* **Ryan LaPointe** - [rrlapointe](https://github.com/rrlapointe)
* **Peter Maida** - [PeterMaidaRobot](https://github.com/PeterMaidaRobot)


## What is it?
The project found within this site is a specification based IDS/IPS with the aim of reducing the spread and impact of IoT botnets by blocking malicious outbound traffic from a compromised device. It is a Major Qualifying Project submitted to the faculty of Worcester Polytechnic Institute in partial fulfillment of the requirements for the degree of Bachelor of Science in Computer Science.
## Why we did it?
As consumer home automation devices are rapidly gaining popularity, more and more examples exist of these devices being compromised at scale to create large botnets. While it would seem obvious that security is a priority for device manufacturers, that is not always the case. This project was design to have the ability to alert the owner of compromised IoT devices so that they can take remedial action. Beyond that, it is able to to alert the manufacturer of compromised IoT devices so that the manufacturer can have increased awareness of vulnerabilities being exploited in the wild.  Finally, we expect that with adoption, it will help prevent further IoT device exploitation’s given the increased awareness on the manufacturer’s behalf.

## How to build and use it

### Middlebox
This system is built with the Rust programming language and the Cargo project manager. The main way that they program will be built, run, and tested is via Cargo.

First change to the middlebox directory that contains the Cargo project and the Cargo.toml file:<br/>
`cd middlebox`

Install the following dependencies with Cargo:<br/>
`cargo install diesel_cli --no-default-features --features postgres`

In order to build the program from the terminal:<br/>
`cargo build`

In order to build and then run the program from the terminal:<br/>
`cargo run`

In order to build and run the unit and integration tests from the terminal:<br/>
`cargo test`

If you want to have different output printed to the screen export either INFO, DEBUG, or ERROR to the RUST_LOG environment variable.

* The INFO flag will print information about the threads entering and leaving their loops and if a new policy has been applied for a client.<br/>
`export RUST_LOG=INFO`

* The DEBUG flag will print infromation about the amount of bytes recieved by the client, if the signature is verified, data being sent between threads, and data being pulled from the datastore.<br/>
`export RUST_LOG=DEBUG`

* The ERROR flag will print any errors that do not stop the program execution from running.<br/>
`export RUST_LOG=ERROR`

More information about using Cargo can be found at [https://doc.rust-lang.org/cargo/index.html](https://doc.rust-lang.org/cargo/index.html)


### Client
As the client did not have as many performance requirements as the middlebox, it was decided to build the project in Python 3, as there is a lot of library support for performing a variety of tasks. The first thing needed to get the client up and running is Python3. Real Python has outlined a simple guide to install it on a variety of operating systems at [https://realpython.com/installing-python/](https://realpython.com/installing-python/). Once Python 3 is installed, the user needs to create  a virtual environment. Virtual environments allows users to create their own isolated site directories, such that only the required packages and nothing more needs to exist in the environment. To create a virtual environment, type `python3 -m venv /path/to/new/virtual/environment`. Pip is the prefered installer program. Starting with Python 3.4, it is included by default with the Python installation. Once the environment is created, you need to source the environment. This is a way of loading into the enviorment with the required packages such that when you run the python scripts they will work as desired. this is done by typing `$source environment_name/bin/activate`. To unsource the environment, type `$deactivate`. While in the environment, there are a few packages that need to be installed. To make it simple for the user, we make a requirements.txt file that will tell pip which packages to install. To install the packages, type `pip install -r requirement.txt`. As a note, pip is installed by default after Python version 3.4. Once this is done, running the project can be done by typing `python3 -m iot_client_sim`. This will run the __main__.py file which will then trigger trigger the normal\_traffic.py and send\_advertisement.py on regular intervals.

### Tests

To run the smoke tests, create a virtual environment and run `pip install -r requirements.txt` and then `pytest`, both while in the `tests/` directory.

## How it works

### Middlebox

#### Advertisements

The advertisement thread runs from the advertisement.rs file. This thread will infinitely loop, recieveing connections and processing them. The code will check if it has recieved a policy with a valid certificate. If it has, the new policy will be added to the database or updated if the reviision number of the new polcicy is higher than the current policy. If the certificate is not valid or the revision nunmber is lower, the new policy will not be added to the database. Both of these two conditions together protect our system from Man-in-the-middle attacks and replay attacks.

#### Firewall

Iptables and nfqueue are used by the firewall to assist with processing packets. The firewall evaluates each new network flow against the applicable policies and decides to pass, drop, or reject the flow.

#### TLS Interception

The middlebox receives TLS key information in SSLKEYLOGFILE format transmitted to it by the client. The key information is fed to Tshark, which decrypts the HTTPS connection. Tshark tells our middlebox code the contents of the HTTP host header, which is then used for enforcing firewall policy.

The decryption process occurs periodically. Every ?? milliseconds any HTTP host headers from the last ?? milliseconds are obtained. Ideally we would want the decryption to happen in real-time but we did what we could for this proof-of-concept project.

#### Centralized Policy Manager

When the middlebox receives an advertisement from a client, it uses the tags contained in the advertisement to fetch managed policies from an FTP server. The firewall applies both the policies sent by the client in its advertisement and any centrally-managed policies that apply to the tags sent by the client.

### Client
The client’s primary goal is to simulate normal traffic and send advertisements.  Utilizing  the  Tox  tool  with  thePyLint and Flake8 plugins, the team was able to ensure code is standardized and will run without problems from multiple developers. Within the client code, there are three primary files that get run. The first file is normal\_traffic.py. This file is responsible for generating normal network traffic outlined below. The second file is send\_advertisement.py. This file is responsible for sending the multicasted advertisement. The final file is a \_\_main\_\_.py. This file is responsible for running both normal\_traffic.py and send\_advertisement.py. It first creates a thread for each file. After that, it will run normal\_traffic.py on every three second interval and send\_advertisement.py on every five second interval. This timing allows for the there to be non-collisions, near collisions, and collisions within the network.
